{

 ****************************************************************************
 ***                   ABACUS'  "PC UNDERGROUND"                          ***
 ***                  ================================                    ***
 ***                                                                      ***
 ***                      Sample program MOD386                           ***
 ***                                                                      ***
 ***  The program demonstrates the use of the MOD_SB unit. You can        ***
 ***  play back 4-channel and 8-channel MOD files. Press "p" during       ***
 ***  playback to view the remaining processor performance. This is       ***
 ***  determined by measuring the time that is available before a retrace.***
 ***                                                                      ***
 *** Author           : Boris Bertelsons  (InspirE)                       ***
 *** Filename         : MOD386.PAS                                        ***
 *** Last update      : 04/04/94                                          ***
 *** Version          : 2.0                                               ***
 *** Compiler         : Turbo Pascal 6.0 and above                        ***
 ****************************************************************************

}



uses crt,dos,mod_sb,variab,design;

{ $define polling}
{
Usually output takes place via the timer interrupt. However, if you have to 
synchronize with the horizontal retrace, then you must use the polling method.
This method doesn't calculate the sound periodically, but rather, calculates it 
when it is time. Polling is somewhat slower and results in tonal losses. In addition,
it can lead to problems with 8-channel MODs. Therefore, use it only when it
can't be avoided.
}

type
  t = record                      { For direct screen output        }
       c : char;
       a : byte;
     end;

const Nummods : byte = 0;
      repeatmode : boolean = false;

var gi : integer;
    my_modname : string;
    stapo,stinc : integer;
    ch,dch : char;
    next_song : integer;
    effects : array[1..4] of effect_type;
    Modd : array[1..10] of string;

procedure Scale_box;
var li : integer;
begin;
  textcolor(1);
  textbackground(black);
  clrscr;
  write(' MOD386  Version 2.0, (c) 1995 ABACUS/1994 DATA BECKER',
	      ' Coding: Boris Bertelsons (InspirE)');
  textcolor(lightblue);
  for li := 1 to 10 do begin;
    gotoxy(2,li+4);
    write(li:2,'. ',instnames[li]);
    gotoxy(28,li+4);
    write(li+10:2,'. ',instnames[li+10]);
    gotoxy(54,li+4);
    write(li+20:2,'. ',instnames[li+20]);
  end;
  if Voices = 4 then begin;
  textcolor(black);
  textbackground(7);
  writexy(02,16,''+
	              'ͻ');
  writexy(02,17,'                                            '+
	              '                               ');
  writexy(02,18,'                                            '+
	              '                               ');
  writexy(02,19,'                                            '+
              	'                               ');
  writexy(02,20,'                                            '+
	              '                               ');
  writexy(02,21,''+
	              'ͼ');
  gotoxy(1,23);
  end else begin;
  textcolor(black);
  textbackground(7);
  writexy(02,16,''+
	              'ͻ');
  writexy(02,17,'                                            '+
	              '                               ');
  writexy(02,18,'                                            '+
	              '                               ');
  writexy(02,19,'                                            '+
	              '                               ');
  writexy(02,20,'                                            '+
	              '                               ');
  writexy(02,21,'                                            '+
	              '                               ');
  writexy(02,22,'                                            '+
 	              '                               ');
  writexy(02,23,'                                            '+
	              '                               ');
  writexy(02,24,'                                            '+
	              '                               ');
  writexy(02,25,''+
	              'ͼ');
  gotoxy(1,23);
  end;
  textbackground(black);
  textcolor(lightblue);
  writexy(47,2,'Volume: ');
  textcolor(lightcyan);
  write(Mastervolume:2);
  textcolor(lightblue);
  textbackground(black);
  writexy(58,2,'Balance ');
  textcolor(14);
  writexy(66,2,'');
  textcolor(4);
  writexy(78-Balance DIV 2,2,'');
  textcolor(lightblue);
  textbackground(black);
  writexy(36,2,'Filter');
  textcolor(lightcyan);
  write(' OFF');
end;

procedure Scale;
var li,lj : integer;
    screen : array[1..50,1..80] of t absolute $B800:$0000;
    secu : string[3];
begin;
  textcolor(lightblue);
  textbackground(black);
  str(Runsec,secu);
  if runsec < 10 then secu := '0'+secu;
  gotoxy(2,3);
  write('Songname : ');
  textcolor(lightcyan);
  write(Songname);
  gotoxy(34,3);
  textcolor(lightblue);
  write('Frequency : ');
  textcolor(lightcyan);
  write(Sampling_Frequency:5);
  textcolor(lightblue);
  write(' KHz   Runtime : ');
  textcolor(lightcyan);
  write(Runmin:2,':',secu);
  gotoxy(2,4);
  textcolor(lightblue);
  write('Pattern No.:');
  textcolor(lightcyan);
  write(Song[mlj]:3);
  textcolor(lightblue);
  write('    Pattern :');
  textcolor(lightcyan);
  write(mlj:3,'/',Songlength:3);
  textcolor(lightblue);
  write('    Line :');
  textcolor(lightcyan);
  write(mli:3);
  textcolor(lightblue);
  write('    Speed : ');
  textcolor(lightcyan);
  write(Playspeed:3,'/128');
  gotoxy(2,2);
  textcolor(lightblue);
  write('Memory Free : ');
  textcolor(lightcyan);
  write(Maxavail:6,' KB');
  textcolor(black);
  textbackground(7);
  for li := 1 to Voices do
    if In_St[li] <> 0 then writexy(4,16+li,Instnames[In_St[li]]);

  for lj := 1 to Voices do begin;
    for li := 1 to 16 do begin;
      if (Notes_Struck[lj] div 10) > li then
        begin;
          screen[16 +lj,29+li].c := '';
          screen[16 +lj,29+li].a := 114;
        end else begin;
          screen[16 +lj,29+li].c := '';
          screen[16 +lj,29+li].a := 112;
        end;
    end;
    for li := 16  to 32 do begin;
      if (Notes_Struck[lj] div 10) > li then
        begin;
          screen[16 +lj,29+li].c := '';
          screen[16 +lj,29+li].a := 126;
        end else begin;
          screen[16 +lj,29+li].c := '';
          screen[16 +lj,29+li].a := 112;
        end;
    end;
    for li := 33 to 48 do begin;
      if (Notes_Struck[lj] div 10) > li then
        begin;
          screen[16 +lj,29+li].c := '';
          screen[16 +lj,29+li].a := 116;
        end else begin;
          screen[16 +lj,29+li].c := '';
          screen[16 +lj,29+li].a := 112;
        end;
    end;
  end;
end;

var retraceincs : word;
    systemspeed : longint;
    modspeed    : longint;

procedure test_waitretrace;
begin;
retraceincs := 0;
  asm
    MOV   DX,03dAh
@WD_R:
    inc word ptr retraceincs
    IN    AL,DX
    TEST  AL,8d
    JZ    @WD_R
@WD_D:
    inc word ptr retraceincs
    IN    AL,DX
    TEST  AL,8d
    JNZ   @WD_D
  end;
end;

procedure test_systemspeed;
var li : integer;
begin;
  writeln;
  writeln('Testing system speed, please wait ...');
  writeln;
  test_waitretrace;
  systemspeed := 0;
  for li := 1 to 70 do begin;
    test_waitretrace;
    systemspeed := systemspeed+retraceincs;
  end;
end;

procedure test_modspeed;
var li : integer;
begin;
  writeln;
  writeln('Testing MOD speed, please wait ...');
  writeln;
  test_waitretrace;
  modspeed := 0;
  for li := 1 to 210 do begin;
    test_waitretrace;
    modspeed := modspeed+retraceincs;
  end;
  modspeed := modspeed div 3;
end;

procedure write_performance;
begin;
  writeln;
  writeln('Remain :',(modspeed*100/systemspeed):6:2,' % Processor performance');
  writeln;
  writeln;
  writeln;
  write('Press Enter to continue ...');
  readln;
end;

procedure Play_the_Mod(s : string);
var h : byte;
    error : integer;
    li : integer;
begin;
{ if not SB16Detected then} Reset_Sb16;
 mod_SetSpeed(66);
 mod_Samplefreq(Samfreq);
 dsp_rdy_sb16 := true;
 error := load_modfile(s,AUTO,AUTO,Samfreq);
 if error <> 0 then begin;
   clrscr;
   writeln('Error loading MOD file ! ');
   if error = -1 then writeln('File not found !');
   if error = -2 then writeln('Insufficient memory available !');
   halt(0);
 end;
 {$ifdef polling}
   start_polling;
 {$else}
   periodic_on;   { Switches on periodic playback }
 {$endif}

 Scale_box;
 ch := #255;
 while not (ch=#27) and not (upcase(ch)='X')
   and not (upcase(ch)='N') do begin;
  {$ifdef polling}
    mod_waitretrace(15);
  {$endif}
  Scale;
  if keypressed then ch := readkey;
  case ch of
    #0  : begin;
            dch := readkey;
            case dch of
              #61 : begin; { F3 }
                       if Mastervolume > 0 then dec(Mastervolume);
                       Set_Volume(Mastervolume);
                       textbackground(black);
                       textcolor(lightblue);
                       writexy(47,2,'Volume: ');
                       textcolor(lightcyan);
                       write(Mastervolume:2);
                       ch := #255;
                     end;
              #62 : begin; { F4 }
                       if Mastervolume < 31 then inc(Mastervolume);
                       Set_Volume(Mastervolume);
                       textbackground(black);
                       textcolor(lightblue);
                       writexy(47,2,'Volume: ');
                       textcolor(lightcyan);
                       write(Mastervolume:2);
                       ch := #255;
                     end;
              #63 : begin; { F5 }
                       if Balance > 0 then dec(Balance);
                       Set_Balance(Balance);
                       textcolor(lightblue);
                       textbackground(black);
                       writexy(58,2,'Balance ');
                       textcolor(14);
                       writexy(66,2,'');
                       textcolor(4);
                       writexy(78-Balance DIV 2,2,'');
                       ch := #255;
                     end;
              #64 : begin; { F6 }
                       if Balance < 24 then inc(Balance);
                       Set_Balance(Balance);
                       textcolor(lightblue);
                       textbackground(black);
                       writexy(58,2,'Balance ');
                       textcolor(14);
                       writexy(66,2,'');
                       textcolor(4);
                       writexy(78-Balance DIV 2,2,'');
                       ch := #255;
                     end;
              else begin;
                     ch := #255;
                   end;
             end;
          end;
    '6' : begin;
            inc(mli);
            ch := #255;
          end;
    'P',
    'p' : begin;
            textcolor(7);
            textbackground(black);
            clrscr;
            test_modspeed;
            write_performance;
            Scale_box;
            ch := #255;
          end;
    'f' : begin;
            filter_active := not filter_active;
            if filter_active then begin;
						  filter_on;
              textcolor(lightblue);
              textbackground(black);
              writexy(36,2,'Filter');
              textcolor(lightcyan);
              write(' ON ');
            end else begin;
						  filter_mid;
              textcolor(lightblue);
              textbackground(black);
              writexy(36,2,'Filter');
              textcolor(lightcyan);
              write(' OFF');
            end;
            ch := #255;
          end;
    '4' : begin;
            if mli > 0 then
              dec(mli)
            else begin;
              if mlj > 0 then begin;
                dec(mlj);
                mli := 63
              end else begin;
                mli := 0;
                mlj := 0;
              end;
            end;
            ch := #255;
           end;
    '3' : begin;
            mli := 0;
            inc(mlj);
            ch := #255;
          end;
    '1' : begin;
            if mlj > 0 then begin;
              dec(mlj);
              mli := 0;
            end;
            ch := #255;
           end;
    'N',
    'n' : begin;
            next_song := 1;
          end;
    'x' : begin;
            next_song := 255;
          end;
    #27 : begin;
            next_song := 255;
          end;
    else begin;
           ch := #255;
    end;
   end;
 end;
 outfading := true;
 while outvolume > 1 do begin;
   Scale;
 end;
 {$ifndef polling}
   periodic_off;
 {$endif}
 end_mod;
{ if not SB16Detected then} Reset_Sb16;
end;

procedure Write_Helptext;
begin;
  textcolor(lightgray);
  textbackground(black);
  clrscr;
  writeln(' MOD386  Version 2.0, (c) 1995 ABACUS/1994 DATA BECKER',
	      ' Coding: Boris Bertelsons (InspirE)');
  writeln;
  writeln(' Usage: Mod386 <Filename[.MOD]> [options]');
  writeln;
  writeln(' Options are:');
	writeln(' -H        : This screen');
  writeln(' -In       : Use interrupt n');
	writeln(' -Dn       : Use DMA channel n');
  writeln(' -Pxxx     : Use base address xxx');
  writeln(' -Snn      : Select sample rate in KHz. Allowed: ',
	        '8,10,13,16,22');
  writeln(' -r        : Enable repeat select mode');
	writeln(' -sb       : SB16 not detected');
  writeln(' <name>    : extra .MOD file, plays in random',
	        ' sequence');
  writeln;
  writeln;
  writeln('                    - key for more -');
  writeln;
  repeat until keypressed; readkey;
  clrscr;
  writeln(' MOD386  Version 2.0, (c) 1995 ABACUS/1994 DATA BECKER',
	      ' Coding: Boris Bertelsons (InspirE)');
  writeln;
  writeln(' Keyboard layout during playback ');
  writeln;
  writeln('  F : Toggles X-bass filter On/Off');
  writeln(' F3 : Volume softer      F4 : Volume louder ');
  writeln(' F5 : Balance to left     F6 : Balance to right');
  writeln('  1 : One pattern back       3 : One pattern forward');
  writeln('  4 : One line back        6 : One line forward');
  writeln('  n : Next file        esc,X : Exit');
  writeln('  p : Remaining system performance');
  writeln;
  Cursor_On;
  halt(0);
end;

procedure interpret_commandline;
var cs,hs : string;
    li,code : integer;
    samplefr : word;
    Datnm : boolean;
begin;
  for li := 1 to 10 do begin;
    cs := paramstr(li);
    Datnm := true;
    { Help requested ? }
    if (pos('-h',cs) <> 0) or (pos('/h',cs) <> 0) or
       (pos('-H',cs) <> 0) or (pos('/H',cs) <> 0) or
       (pos('-?',cs) <> 0) or (pos('/?',cs) <> 0) then begin;
       write_Helptext;
      Datnm := false;
    end;
    { Repeatmode ? }
    if (pos('-r',cs) <> 0) or (pos('/r',cs) <> 0) or
       (pos('-R',cs) <> 0) or (pos('/R',cs) <> 0) then begin;
       Repeatmode := true;
      Datnm := false;
    end;
    { Force NO Sb16 ? }
    if (pos('-sb',cs) <> 0) or (pos('/sb',cs) <> 0) or
       (pos('-SB',cs) <> 0) or (pos('/SB',cs) <> 0) then begin;
       force_SB := true;
      Datnm := false;
    end;
    if (pos('-i',cs) <> 0) or (pos('/i',cs) <> 0) or
       (pos('-I',cs) <> 0) or (pos('/I',cs) <> 0) then begin;
      force_irq := true;
      hs := copy(cs,3,length(cs)-2);
      val(hs,dsp_irq,code);
      Datnm := false;
    end;
    { Force DMA ? }
    if (pos('-d',cs) <> 0) or (pos('/d',cs) <> 0) or
       (pos('-D',cs) <> 0) or (pos('/D',cs) <> 0) then begin;
      force_dma := true;
      hs := copy(cs,3,length(cs)-2);
      val(hs,dma_ch,code);
      Datnm := false;
    end;
    { Force Base ? }
    if (pos('-p',cs) <> 0) or (pos('/p',cs) <> 0) or
       (pos('-P',cs) <> 0) or (pos('/P',cs) <> 0) then begin;
       hs := copy(cs,3,length(cs)-2);
       if hs = '200' then dsp_addr := $200;
       if hs = '210' then dsp_addr := $210;
       if hs = '220' then dsp_addr := $220;
       if hs = '230' then dsp_addr := $230;
       if hs = '240' then dsp_addr := $240;
       if hs = '250' then dsp_addr := $250;
       if hs = '260' then dsp_addr := $260;
       if hs = '270' then dsp_addr := $270;
       if hs = '280' then dsp_addr := $280;
       Startport := dsp_addr;
       Endport   := dsp_addr;
      Datnm := false;
    end;
    { Set sample rate ? }
    if (pos('-s',cs) <> 0) or (pos('/s',cs) <> 0) or
       (pos('-S',cs) <> 0) or (pos('/S',cs) <> 0) then begin;
      hs := copy(cs,3,length(cs)-2);
      val(hs,Samplefr,code);
      if Samplefr >= 8000 then Samplefr := Samplefr DIV 1000;
      if Samplefr >= 8 then Samfreq := 8;
      if Samplefr >= 10 then Samfreq := 10;
      if Samplefr >= 13 then Samfreq := 13;
      if Samplefr >= 16 then Samfreq := 16;
      if Samplefr >= 22 then Samfreq := 22;
      Datnm := false;
    end;
    if Datnm then begin;
      if cs <> '' then begin;
        Inc(Nummods);
        Modd[Nummods] := cs;
      end;
    end;
  end;
end;

procedure write_vocmessage;
begin;
  clrscr;
  writexy(10,08,'Warning !  VOC looping relentlessly !!!');
  writexy(10,10,'To exit, press >>  Q  <<');
  writexy(10,14,'If possible, remove Smartdrv, because it is slooooooow !');
  writexy(10,21,'                            E N J O Y');
end;

procedure play_sound(datname : string);
var li : integer;
    ch : char;
begin;
  for li := 1 to length(datname) do
     datname[li] := upcase(Datname[li]);
  if pos('.MOD',datname) <> 0 then begin;
    Play_The_Mod(datname);
    exit;
  end;
  if pos('.VOC',datname) <> 0 then begin;
  repeat
    Reset_Sb16;
    write_vocmessage;
    Init_Voc(datname);
    ch := #0;
    repeat
      if keypressed then ch := readkey;
      if ch = 'p' then begin;
        voc_pause;
        repeat
          ch := readkey;
        until ch = 'c';
        voc_continue;
      end;
    until VOC_READY or (ch = 'n') or (upcase(ch) = 'Q');
    VOC_DONE;
  until upcase(ch) = 'Q';
  end;
end;


begin;
 Samfreq := 22;
 clrscr;
 test_systemspeed;
 cursor_off;
 interpret_commandline;
 if (Nummods = 0) and not repeatmode then begin;
   textcolor(15);
   textbackground(1);
   clrscr;
   Nummods := 1;
   modd[1] := ''+select_file('*.?o?','*.?o?','','Please select MOD file');
   if modd[1] = 'xxxx' then begin;
     clrscr;
		 writeln('You must already have ''as a MOD file !');
     Cursor_on;
     halt(0);
   end;
 end;
 for i := 1 to Nummods do begin;
   if pos('.',modd[i]) = 0 then modd[i] := modd[i]+'.mod';
 end;
 Init_The_Mod;
 stereo := false;
 next_song := random(Nummods)+1;
 textcolor(lightgray);
 textbackground(black);
 write_sbconfig;
 writeln;
 writeln;
 write('                     ENTER  to continue ...');
 readln;
 repeat
   if repeatmode then begin;
     textcolor(15);
     textbackground(1);
     clrscr;
     modd[1] := ''+select_file('*.?o?','*.?o?','','');
     if modd[1] = 'xxxx' then next_song := 255
     else Play_Sound(modd[1]);
   end else
     Play_Sound(modd[next_song]);
   if next_song <> 255 then next_song := random(Nummods)+1;
 until next_song = 255;
 cursor_on;
 textmode(3);
end.
